//---------------------------------------------------------------------------
//
//        This file contains the client side implementation of the
//        sqlite C++ interface for socket communication
//
//---------------------------------------------------------------------------
//
#pragma hdrstop


// This include defines the interface of the exported methods.
// Does NOT mean usage of sqlite shared library.(This why "sqlite.h" instead of <sqlite>)
// (This file is part of the client package because client
// lib does not need full sqlite library installed.)
#include "_sqlite.h"

#include "clientStubs.h"
#include <assert.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

//---------------------------------------------------------------------------
#pragma package(smart_init)

#define CLIENT_ERROR_BASE 1000
#define debug(x) printf(x)


int libinited = 0;

class sqlite
{
  public:
  int handle;
  int flags;
};

/*
** Possible values for the sqlite.flags and or Db.flags fields.
**
** On sqlite.flags, the SQLITE_InTrans value means that we have
** executed a BEGIN.  On Db.flags, SQLITE_InTrans means a statement
** transaction is active on that particular database file.
*/
#define SQLITE_VdbeTrace      0x00000001  /* True to trace VDBE execution */
#define SQLITE_Initialized    0x00000002  /* True after initialization */
#define SQLITE_Interrupt      0x00000004  /* Cancel current operation */
#define SQLITE_InTrans        0x00000008  /* True if in a transaction */
#define SQLITE_InternChanges  0x00000010  /* Uncommitted Hash table changes */
#define SQLITE_FullColNames   0x00000020  /* Show full column names on SELECT */
#define SQLITE_ShortColNames  0x00000040  /* Show short columns names */
#define SQLITE_CountRows      0x00000080  /* Count rows changed by INSERT, */
                                          /*   DELETE, or UPDATE and return */
                                          /*   the count using a callback. */
#define SQLITE_NullCallback   0x00000100  /* Invoke the callback once if the */
                                          /*   result set is empty */
#define SQLITE_ReportTypes    0x00000200  /* Include information on datatypes */
                                          /*   in 4th argument of callback */

                                          
void delete_StrArray(STACK_StrArray &sa)
{
    delete[] sa.items;
    sa.len = 0;
}

class sqlite_vm
{

   public:

   int handle;
   STACK_StrArray colValues;
   STACK_StrArray colNames;
   void clear()
   {
     delete_StrArray(colValues);
     delete_StrArray(colNames);
   }
   sqlite_vm()
   {
    colValues.items = 0;
    colNames.items = 0;
   }
   ~sqlite_vm()
   {
     clear();
   }

};


void sqliteStrRealloc(char **str)
{
  if (str==0 || *str==0) return;
  char *zNew = strdup(*str);
  free(*str);
  (*str) = zNew;
}

/*
** Free memory previously allocated
*/
void sqliteFree(void *p){
  if( p ){
    free(p);
  }
}

// splits the string into head and tail at the first
// occurance of delim character
void split(char* str, char delim, char** head, char** tail)
{
  int len = strlen(str);

  for (int i=0;i<len;i++)
  {
    if (str[i] == delim)
    {
      str[i] = 0;
      *head = str;
      *tail = str+i+1;
    }
  }

}

void parseDatabaseURL(char* URL, char* &dbname)
{
  char* userpass = 0;
  char* serverdb = URL;

  split(URL,'@',&userpass,&serverdb);

  //char* dbname;
  char* serverport;
  split(serverdb,'/',&serverport,&dbname);

  char* server;
  char* port;

  split(serverport,':',&server,&port);

  int  iport;
  sscanf(port,"%d",&iport);

  if (libinited == 1)
    {
      printf("init client stubs\n");
      clientStubs_init(server,iport);
    }
}

//---------------------------------------------------------------------------
//  Exported   methods
//---------------------------------------------------------------------------


void sqlite_freemem(void *p)
{
   sqliteFree(p);
}



sqlite *sqlite_open(const char *zFilename, int mode, char **pzErrMsg)
{
  char *dbname;
  libinited++;

  parseDatabaseURL((char*)zFilename,dbname);

  sqlite *db;
  db = new sqlite;
  int errcode;
  int res = svc_sqlite_open((char*)dbname, mode, &errcode,pzErrMsg,&(db->handle));
  if ((res == SktStack::STACK_OK) && (errcode == SQLITE_OK))
  {
    return db;
  }
  else
  {
     delete db;
     return 0;
  }
}

void sqlite_close(sqlite *db)
{
   svc_sqlite_close(db->handle);
   delete db;
   libinited--;
   if (libinited == 0)
   {
    clientStubs_deinit();
   }
}


/*
** Compile a single statement of SQL into a virtual machine.  Return one
** of the SQLITE_ success/failure codes.  Also write an error message into
** memory obtained from malloc() and make *pzErrMsg point to that message.
*/
int sqlite_compile(
  sqlite *db,                 /* The database on which the SQL executes */
  const char *zSql,           /* The SQL to be executed */
  const char **pzTail,        /* OUT: Next statement after the first */
  sqlite_vm **ppVm,           /* OUT: The virtual machine */
  char **pzErrMsg             /* OUT: Write error messages here */
)
{
  int hvm;
  int errcode;
  unsigned int tailIndex;

  int res = svc_sqlite_compile(db->handle,(char*)zSql,(int*)&tailIndex,&hvm,&errcode,pzErrMsg);
  if ((res == SktStack::STACK_OK))
  {
    if (tailIndex)
      {
	(*pzTail) = (char*)((unsigned int)zSql + tailIndex - 1);
      }
    else
      {
	(*pzTail) = 0;
      }

    if (errcode != SQLITE_OK)
     {
      *ppVm = 0;
     }
    else
    {
      sqlite_vm *vm = new sqlite_vm;
      vm->handle = hvm;
      *ppVm = vm;
    }
    return errcode;
  }
  else
  {
    *ppVm = 0;
    return CLIENT_ERROR_BASE+res;
  }
}



int sqlite_step(
  sqlite_vm *pVm,              /* The virtual machine to execute */
  int *pN,                     /* OUT: Number of columns in result */
  const char ***pazValue,      /* OUT: Column data */
  const char ***pazColName     /* OUT: Column names and datatypes */
)
{
  int errcode;
  pVm->clear();
  int res = svc_sqlite_step(pVm->handle,&(pVm->colValues),&(pVm->colNames),&errcode);
  if ((res == SktStack::STACK_OK))
  {
    *pN = pVm->colValues.len;
    *pazValue = (const char**)pVm->colValues.items;
    *pazColName = (const char**)pVm->colNames.items;
    return errcode;
  }
  else
  {
    return CLIENT_ERROR_BASE+res;
  }
}

int sqlite_finalize(sqlite_vm* vm, char **pzErrMsg)
{
  if (vm == 0) return SQLITE_OK;
  int errcode;

  int res = svc_sqlite_finalize(vm->handle,&errcode,pzErrMsg);
  if ((res == SktStack::STACK_OK))
  {
    delete vm;
    return errcode;
  }
  else
  {
    return CLIENT_ERROR_BASE+res;
  }
}

/*
** This routine deletes the virtual machine, writes any error message to
** *pzErrMsg and returns an SQLite return code in the same way as the
** sqlite_finalize() function.
**
** Additionally, if ppVm is not NULL, *ppVm is left pointing to a new virtual
** machine loaded with the compiled version of the original query ready for
** execution.
**
** If sqlite_reset() returns SQLITE_SCHEMA, then *ppVm is set to NULL.
**
******* THIS IS AN EXPERIMENTAL API AND IS SUBJECT TO CHANGE ******
*/
int sqlite_reset(sqlite_vm* vm, char **pzErrMsg)
{
  if (vm == 0) return SQLITE_OK;
  int errcode;
  int res = svc_sqlite_reset(vm->handle,&errcode,pzErrMsg);
  if ((res == SktStack::STACK_OK))
  {
    delete vm;
    return errcode;
  }
  else
  {
    return CLIENT_ERROR_BASE+res;
  }
}

int sqlite_changes(sqlite* db)
{
  int changes;
  int res = svc_sqlite_changes(db->handle,&changes);
  if (res == SktStack::STACK_OK)
   return changes;
  else
   return -1;
}

int sqlite_last_insert_rowid(sqlite* db)
{
  int rowid;
  int res = svc_sqlite_last_insert_rowid(db->handle,&rowid);
  if (res == SktStack::STACK_OK)
   return rowid;
  else
   return -1;
}


/* This function causes any pending database operation to abort and
** return at its earliest opportunity.  This routine is typically
** called in response to a user action such as pressing "Cancel"
** or Ctrl-C where the user wants a long query operation to halt
** immediately.
*/
void sqlite_interrupt(sqlite* db)
{
   svc_sqlite_interrupt(db->handle);
}


const char *sqlite_libversion(void)
{
  char* version;
  int res = svc_sqlite_libversion(&version);
  if (res == SktStack::STACK_OK)
   return version;
  else
   return 0;
}

const char *sqlite_libencoding(void)
{
  char* encoding;
  int res = svc_sqlite_libencoding(&encoding);
  if (res == SktStack::STACK_OK)
   return encoding;
  else
   return 0;
}

int sqlite_complete(const char *sql)
{
   //currently no right implementation
   return 1;
}

/*
** Register a function that is called at every invocation of sqlite_exec()
** or sqlite_compile().  This function can be used (for example) to generate
** a log file of all SQL executed against a database.
*/
void *sqlite_trace(sqlite*, void(*xTrace)(void*,const char*), void*)
{
   printf("sqlite_trace is not supported in client\n");
   return 0;
}


int sqlite_set_authorizer(
  sqlite*,
  int (*xAuth)(void*,int,const char*,const char*,const char*,const char*),
  void *pUserData
)
{
  printf("sqlite_set_authorizer is not supported in client\n");
  return 0;
}



/*
** Use the following routines to create new user-defined functions.  See
** the documentation for details.
*/
int sqlite_create_function(
  sqlite*,                  /* Database where the new function is registered */
  const char *zName,        /* Name of the new function */
  int nArg,                 /* Number of arguments.  -1 means any number */
  void (*xFunc)(sqlite_func*,int,const char**),  /* C code to implement */
  void *pUserData           /* Available via the sqlite_user_data() call */
)
{
  return 0;
}

int sqlite_create_aggregate(
  sqlite*,                  /* Database where the new function is registered */
  const char *zName,        /* Name of the function */
  int nArg,                 /* Number of arguments */
  void (*xStep)(sqlite_func*,int,const char**), /* Called for each row */
  void (*xFinalize)(sqlite_func*),       /* Called once to get final result */
  void *pUserData           /* Available via the sqlite_user_data() call */
)
{
   return 0;
}

/*
** Use the following routine to define the datatype returned by a
** user-defined function.  The second argument can be one of the
** constants SQLITE_NUMERIC, SQLITE_TEXT, or SQLITE_ARGS or it
** can be an integer greater than or equal to zero.  The datatype
** will be numeric or text (the only two types supported) if the
** argument is SQLITE_NUMERIC or SQLITE_TEXT.  If the argument is
** SQLITE_ARGS, then the datatype is numeric if any argument to the
** function is numeric and is text otherwise.  If the second argument
** is an integer, then the datatype of the result is the same as the
** parameter to the function that corresponds to that integer.
*/
int sqlite_function_type(
  sqlite *db,               /* The database there the function is registered */
  const char *zName,        /* Name of the function */
  int datatype              /* The datatype for this function */
)
{
   return 0;
}
#define SQLITE_NUMERIC     (-1)
#define SQLITE_TEXT        (-2)
#define SQLITE_ARGS        (-3)

/*
** The user function implementations call one of the following four routines
** in order to return their results.  The first parameter to each of these
** routines is a copy of the first argument to xFunc() or xFinialize().
** The second parameter to these routines is the result to be returned.
** A NULL can be passed as the second parameter to sqlite_set_result_string()
** in order to return a NULL result.
**
** The 3rd argument to _string and _error is the number of characters to
** take from the string.  If this argument is negative, then all characters
** up to and including the first '\000' are used.
**
** The sqlite_set_result_string() function allocates a buffer to hold the
** result and returns a pointer to this buffer.  The calling routine
** (that is, the implmentation of a user function) can alter the content
** of this buffer if desired.
*/
char *sqlite_set_result_string(sqlite_func*,const char*,int)
{
   return 0;
}
void sqlite_set_result_int(sqlite_func*,int)
{

}
void sqlite_set_result_double(sqlite_func*,double)
{

}
void sqlite_set_result_error(sqlite_func*,const char*,int)
{

}

/*
** The pUserData parameter to the sqlite_create_function() and
** sqlite_create_aggregate() routines used to register user functions
** is available to the implementation of the function using this
** call.
*/
void *sqlite_user_data(sqlite_func*)
{
  return 0;
}

/*
** Aggregate functions use the following routine to allocate
** a structure for storing their state.  The first time this routine
** is called for a particular aggregate, a new structure of size nBytes
** is allocated, zeroed, and returned.  On subsequent calls (for the
** same aggregate instance) the same buffer is returned.  The implementation
** of the aggregate can use the returned buffer to accumulate data.
**
** The buffer allocated is freed automatically be SQLite.
*/
void *sqlite_aggregate_context(sqlite_func*, int nBytes)
{
  return 0;
}

/*
** The next routine returns the number of calls to xStep for a particular
** aggregate function instance.  The current call to xStep counts so this
** routine always returns at least 1.
*/
int sqlite_aggregate_count(sqlite_func*)
{
   return 0;
}


/*
** This routine registers a callback with the SQLite library.  The
** callback is invoked for every attempt to access a column of a table
** in the database.  The callback returns SQLITE_OK if access is allowed,
** SQLITE_DENY if the entire SQL statement should be aborted with an error
** and SQLITE_IGNORE if the column should be treated as a NULL value.
*/
int sqlite_set_authorizer(
  sqlite*,
  int (*xAuth)(void*,int,const char*,const char*),
  void *pUserData
)
{
   return 0;
}


/*
** This routine identifies a callback function that is invoked
** whenever an attempt is made to open a database table that is
** currently locked by another process or thread.  If the busy callback
** is NULL, then sqlite_exec() returns SQLITE_BUSY immediately if
** it finds a locked table.  If the busy callback is not NULL, then
** sqlite_exec() invokes the callback with three arguments.  The
** second argument is the name of the locked table and the third
** argument is the number of times the table has been busy.  If the
** busy callback returns 0, then sqlite_exec() immediately returns
** SQLITE_BUSY.  If the callback returns non-zero, then sqlite_exec()
** tries to open the table again and the cycle repeats.
**
** The default busy callback is NULL.
**
** Sqlite is re-entrant, so the busy handler may start a new query. 
** (It is not clear why anyone would every want to do this, but it
** is allowed, in theory.)  But the busy handler may not close the
** database.  Closing the database from a busy handler will delete 
** data structures out from under the executing query and will 
** probably result in a coredump.
*/
void sqlite_busy_handler(sqlite*, int(*)(void*,const char*,int), void*)
{

}

/*
** This routine sets a busy handler that sleeps for a while when a
** table is locked.  The handler will sleep multiple times until 
** at least "ms" milleseconds of sleeping have been done.  After
** "ms" milleseconds of sleeping, the handler returns 0 which
** causes sqlite_exec() to return SQLITE_BUSY.
**
** Calling this routine with an argument less than or equal to zero
** turns off all busy handlers.
*/
void sqlite_busy_timeout(sqlite*, int ms)
{

}

/*
** Return a static string that describes the kind of error specified in the
** argument.
*/
const char *sqlite_error_string(int rc){
  const char *z;
  switch( rc ){
    case SQLITE_OK:         z = "not an error";                          break;
    case SQLITE_ERROR:      z = "SQL logic error or missing database";   break;
    case SQLITE_INTERNAL:   z = "internal SQLite implementation flaw";   break;
    case SQLITE_PERM:       z = "access permission denied";              break;
    case SQLITE_ABORT:      z = "callback requested query abort";        break;
    case SQLITE_BUSY:       z = "database is locked";                    break;
    case SQLITE_LOCKED:     z = "database table is locked";              break;
    case SQLITE_NOMEM:      z = "out of memory";                         break;
    case SQLITE_READONLY:   z = "attempt to write a readonly database";  break;
    case SQLITE_INTERRUPT:  z = "interrupted";                           break;
    case SQLITE_IOERR:      z = "disk I/O error";                        break;
    case SQLITE_CORRUPT:    z = "database disk image is malformed";      break;
    case SQLITE_NOTFOUND:   z = "table or record not found";             break;
    case SQLITE_FULL:       z = "database is full";                      break;
    case SQLITE_CANTOPEN:   z = "unable to open database file";          break;
    case SQLITE_PROTOCOL:   z = "database locking protocol failure";     break;
    case SQLITE_EMPTY:      z = "table contains no data";                break;
    case SQLITE_SCHEMA:     z = "database schema has changed";           break;
    case SQLITE_TOOBIG:     z = "too much data for one table row";       break;
    case SQLITE_CONSTRAINT: z = "constraint failed";                     break;
    case SQLITE_MISMATCH:   z = "datatype mismatch";                     break;
    case SQLITE_MISUSE:     z = "library routine called out of sequence";break;
    case SQLITE_NOLFS:      z = "kernel lacks large file support";       break;
    case SQLITE_AUTH:       z = "authorization denied";                  break;
    case SQLITE_FORMAT:     z = "auxiliary database format error";       break;
    case SQLITE_RANGE:      z = "bind index out of range";               break;
    case SQLITE_NOTADB:     z = "file is encrypted or is not a database";break;
    default:                z = "unknown error";                         break;
  }
  return z;
}


//---------------------------------------------------------------------------
//    The implementation of the methods below is copied from sqlite sources.
//    As long as methods above are implemented correctly, the following
//    methods should also produce correct results.
//---------------------------------------------------------------------------



/*
** Execute SQL code.  Return one of the SQLITE_ success/failure
** codes.  Also write an error message into memory obtained from
** malloc() and make *pzErrMsg point to that message.
**
** If the SQL is a query, then for each row in the query result
** the xCallback() function is called.  pArg becomes the first
** argument to xCallback().  If xCallback=NULL then no callback
** is invoked, even for queries.
*/
int sqlite_exec(
  sqlite *db,                 /* The database on which the SQL executes */
  const char *zSql,           /* The SQL to be executed */
  sqlite_callback xCallback,  /* Invoke this callback routine */
  void *pArg,                 /* First argument to xCallback() */
  char **pzErrMsg             /* Write error messages here */
){
  int rc = SQLITE_OK;
  const char *zLeftover;
  sqlite_vm *pVm;
  int nRetry = 0;
  int nChange = 0;
  int nCallback;

  if( zSql==0 ) return SQLITE_OK;
  while( rc==SQLITE_OK && zSql[0] ){
    pVm = 0;
    rc = sqlite_compile(db, zSql, &zLeftover, &pVm, pzErrMsg);
    if( rc!=SQLITE_OK ){
//      assert( pVm==0 || sqlite_malloc_failed );
      return rc;
    }
    if( pVm==0 ){
      /* This happens if the zSql input contained only whitespace */
      break;
    }
//    db->nChange += nChange;
    nCallback = 0;
    while(1){
      int nArg;
      char **azArg, **azCol;
      rc = sqlite_step(pVm, &nArg, (const char***)&azArg,(const char***)&azCol);
      if( rc==SQLITE_ROW ){
        if( xCallback!=0 && xCallback(pArg, nArg, azArg, azCol) ){
          sqlite_finalize(pVm, 0);
          return SQLITE_ABORT;
        }
        nCallback++;
      }else{
        if( rc==SQLITE_DONE && nCallback==0
          && (db->flags & SQLITE_NullCallback)!=0 && xCallback!=0 ){
          xCallback(pArg, nArg, azArg, azCol);
        }
        rc = sqlite_finalize(pVm, pzErrMsg);
        if( rc==SQLITE_SCHEMA && nRetry<2 ){
          nRetry++;
          rc = SQLITE_OK;
          break;
        }
/*        if( db->pVdbe==0 ){
          nChange = db->nChange;
        }
*/        
        nRetry = 0;
        zSql = zLeftover;
        while( isspace(zSql[0]) ) zSql++;
        break;
      }
    }
  }
  return rc;
}


/*
** This structure is used to pass data from sqlite_get_table() through
** to the callback function is uses to build the result.
*/
typedef struct TabResult {
  char **azResult;
  char *zErrMsg;
  int nResult;
  int nAlloc;
  int nRow;
  int nColumn;
  int nData;
  int rc;
} TabResult;

/*
** This routine is called once for each row in the result table.  Its job
** is to fill in the TabResult structure appropriately, allocating new
** memory as necessary.
*/
static int sqlite_get_table_cb(void *pArg, int nCol, char **argv, char **colv){
  TabResult *p = (TabResult*)pArg;
  int need;
  int i;
  char *z;

  //printf("callback row=%d\n",p->nRow);
  /* Make sure there is enough space in p->azResult to hold everything
  ** we need to remember from this invocation of the callback.
  */
  if( p->nRow==0 && argv!=0 ){
    need = nCol*2;
  }else{
    need = nCol;
  }
  if( p->nData + need >= p->nAlloc ){
    char **azNew;
    p->nAlloc = p->nAlloc*2 + need + 1;
   // debug("try realloc\n");
    azNew = (char**) realloc( p->azResult, sizeof(char*)*p->nAlloc );
    if( azNew==0 ){
      p->rc = SQLITE_NOMEM;
      return 1;
    }
    p->azResult = azNew;
  }

  /* If this is the first row, then generate an extra row containing
  ** the names of all columns.
  */
  if( p->nRow==0 ){
    p->nColumn = nCol;
    for(i=0; i<nCol; i++){
      if( colv[i]==0 ){
        z = 0;
      }else{
//       debug("try malloc\n");
        z = (char*)malloc( strlen(colv[i])+1 );
        if( z==0 ){
          p->rc = SQLITE_NOMEM;
          return 1;
        }
        strcpy(z, colv[i]);
      }
      p->azResult[p->nData++] = z;
    }
  }else if( p->nColumn!=nCol ){
    /*
    sqliteSetString(&p->zErrMsg,
       "sqlite_get_table() called with two or more incompatible queries",
       (char*)0);
    */
    p->rc = SQLITE_ERROR;
    return 1;
  }

  /* Copy over the row data
  */
  if( argv!=0 ){
    for(i=0; i<nCol; i++){
      if( argv[i]==0 ){
        z = 0;
      }else{
        int slen=strlen(argv[i])+1;
      //  printf("slen = %d\n",slen);
      //  debug("try malloc 2\n");
        z = (char*)malloc(slen);
      //  debug("done malloc 2\n");
        if( z==0 ){
          p->rc = SQLITE_NOMEM;
          return 1;
        }
        strcpy(z, argv[i]);
      }
      p->azResult[p->nData++] = z;
    }
    p->nRow++;
  }
  return 0;
}

/*
** Query the database.  But instead of invoking a callback for each row,
** malloc() for space to hold the result and return the entire results
** at the conclusion of the call.
**
** The result that is written to ***pazResult is held in memory obtained
** from malloc().  But the caller cannot free this memory directly.  
** Instead, the entire table should be passed to sqlite_free_table() when
** the calling procedure is finished using it.
*/
int sqlite_get_table(
  sqlite *db,                 /* The database on which the SQL executes */
  const char *zSql,           /* The SQL to be executed */
  char ***pazResult,          /* Write the result table here */
  int *pnRow,                 /* Write the number of rows in the result here */
  int *pnColumn,              /* Write the number of columns of result here */
  char **pzErrMsg             /* Write error messages here */
){
  int rc;
  TabResult res;

  if( pazResult==0 ){ return SQLITE_ERROR; }
  *pazResult = 0;
  if( pnColumn ) *pnColumn = 0;
  if( pnRow ) *pnRow = 0;
  res.zErrMsg = 0;
  res.nResult = 0;
  res.nRow = 0;
  res.nColumn = 0;
  res.nData = 1;
  res.nAlloc = 20;
  res.rc = SQLITE_OK;
  res.azResult = (char**)malloc( sizeof(char*)*res.nAlloc );
  if( res.azResult==0 ){
    return SQLITE_NOMEM;
  }
  res.azResult[0] = 0;
  //debug("about to exec \n");
  rc = sqlite_exec(db, zSql, sqlite_get_table_cb, &res, pzErrMsg);
  if( res.azResult ){
    res.azResult[0] = (char*)res.nData;
  }
  if( rc==SQLITE_ABORT ){
    //debug("aborting\n");
    sqlite_free_table(&res.azResult[1]);
    if( res.zErrMsg ){
      if( pzErrMsg ){
        free(*pzErrMsg);
        *pzErrMsg = res.zErrMsg;
        sqliteStrRealloc(pzErrMsg);
      }else{
        sqliteFree(res.zErrMsg);
      }
    }
    return res.rc;
  }
   sqliteFree(res.zErrMsg);
  if( rc!=SQLITE_OK ){
    //debug("freeing table\n");

    sqlite_free_table(&res.azResult[1]);
    return rc;
  }
  if( res.nAlloc>res.nData ){
    char **azNew;
    azNew = (char**)realloc( res.azResult, sizeof(char*)*(res.nData+1) );
    if( azNew==0 ){
     // debug("freeing table 2\n");
      sqlite_free_table(&res.azResult[1]);
      return SQLITE_NOMEM;
    }
    res.nAlloc = res.nData+1;
    res.azResult = azNew;
  }
  *pazResult = &res.azResult[1];
  if( pnColumn ) *pnColumn = res.nColumn;
  if( pnRow ) *pnRow = res.nRow;
  return rc;
}

/*
** This routine frees the space the sqlite_get_table() malloced.
*/
void sqlite_free_table(
  char **azResult             /* Result returned from from sqlite_get_table() */
){
  if( azResult ){
    int i, n;
    azResult--;
    if( azResult==0 ) return;
    n = (int)azResult[0];
    for(i=1; i<n; i++){ if( azResult[i] ) free(azResult[i]); }
    free(azResult);
  }
}


